#ifndef MARK5B_CONNECTION_H
#define MARK5B_CONNECTION_H

#include <su_control/message_defs.h>
#include <su_control/generated_mess_defs.h>
#include <su_control/ndbg_mark5b.h>

#include <sys/time.h>
#include <string>
#include <vector>


class Mark5B_Connection {
 public:
  Mark5B_Connection();
  virtual ~Mark5B_Connection();

  bool connect_socket(const std::string& host); // host should be a ip address or host name

  bool connected();
  int retreive_messages(timeval& max_wait, int su_id); // waits at max max_wait seconds for ANY message from the mark5B, returns the number of messages received 
  // messages are stored in an internal buffer
  // max_wait is undefined afterwards

  
  // returns a pointer to the message if any, else null
  // reponsibility of the caller to delete the message
  template<class Msg_Type> Msg_Type* get_message();
  
  template<class Msg_Type> bool send_message(const Msg_Type& msg);
  
 protected:
  Mark5B_Connection(Mark5B_Connection&); // copying would mess up the internal buffer, so this method is undefined

  char _message_buffer[MAX_MESSAGE_SIZE]; // messages received will be inserted temporary in this buffer
  char _read_buffer[MAX_MESSAGE_SIZE]; // bytes received from the socket are inserted into this buffer for processing
  char* _read_pointer; // current pointer in the buffer; the first byte in the buffer not read by a previous read_value call
  char* _end_of_read_buffer; // the first byte in the buffer that does not contain a byte read by a previous read_chunk call

  char _write_buffer[MAX_MESSAGE_SIZE];
  char* _write_pointer;

  int _client_socket;
  bool _connected;

  // use internal buffer to store messages
  typedef std::vector< void* > Message_Storage;
  Message_Storage _received_messages;

  Message_Storage::iterator find_message(int msg_code);

  // functions to read the bytes from the socket stream into messages

  int read_chunk(int* client_socket, int number_of_bytes); // tries to put number_of_bytes in the read buffer, any bytes not yet used from previous read will be re-used // if not all bytes are use (by read_value) before calling read_chunk again, it will assume no bytes have been used

  void read_value(char* buffer, int number_of_bytes);
  
  template<class Msg_Type> bool receive_msg_content(int* client_socket, Msg_Type* msg, int su_id);

  // functions to write the messages into a (byte) buffer before it is sent
  void write_value(const char* buffer, int number_of_bytes);

#include <su_control/generated_mess_headers.h>
};

// the template definitions
#include <su_control/generated_mess_templates.h>

template<class Msg_Type> bool Mark5B_Connection::send_message(const Msg_Type& msg) {
  NDBG(LOGSTREAM,MESSAGES|CONNECTION,"Sending message of type " << typeid(msg).name() << std::endl << msg << std::endl);
  int out;
  if (_client_socket >= 0) {
    _write_pointer = &_write_buffer[0];
    out = write_msg(msg.bin_head);
    out += write_msg(msg);

    int bytes_written = write(_client_socket,_write_buffer,out);
    if (bytes_written < 0) {
      NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR writing to socket");
      return false;
    }
    if (bytes_written != out) {
      NDBG(LOGSTREAM,WARNING|CONNECTION,"Did not write complete message, bytes sent: " << bytes_written << " of " <<  out);
      return false;
    }
    return true;
  }
  else {
    return false;
  }
}

template<class Msg_Type> bool Mark5B_Connection::receive_msg_content(int* client_socket, Msg_Type* msg, int su_id) {
  if (receive_msg(client_socket, msg)) {
    if (msg->su_id == su_id) {
      _received_messages.push_back(new Msg_Type(*msg));
      return true;
    }
    else {
      NDBG(LOGSTREAM,CONNECTION|WARNING,"Received message addressed to " << (int)msg->su_id << " this id " << su_id << std::endl);
      return false;
    }
  }
  else {
      NDBG(LOGSTREAM,CONNECTION|WARNING,"Could not read complete message from socket for message type " << typeid(Msg_Type).name() << std::endl);
    return false;
  }
}


#endif
