#ifndef MARK5B_CONTROL_H
#define MARK5B_CONTROL_H

#include <su_control/su_controller.h>
#include <su_control/su_defs.h>
#include <su_control/transp.h>
#include <su_control/su_cmd.h>
#include <su_control/mark5b_connection.h>
#include <su_control/du_status.h>
#include <su_control/scan_iterator.h>
#include <su_control/init_proc.h>

#include <controllers/thread.h>
#include <controllers/state_machine.h>

#include <jccs_util/defs.h>

#include <herbert/signal_flag.h>
#include <herbert/nicetimer.h>

#include <prep_job/polynomials.h>
#include <prep_job/polys_db.h>
#include <hw_sw_setup_db/hw_sw_setup_db.h>

#include <exception>

//#define MK5B_TEST

// the domino rot2tot offset depends on certain buffer sizes in the domino, these values are
const int MBB_SIZE = 128204800;          // circular buffer size (in 32 bit words) 
const int MBB_CHUNK_FACTOR = 500800;     // size of write from buffer to FPDP (bytes), will be multiplied by bytes/us
const double DOM_BUFF_MB = 256.0;        // size of circular buffer within the DOM (MB)
const double FIFO_LAT = 1.0;             // latency (i.e. target fill size) of SS fifo (s)

const int EVLBI_LOAD_MODEL_DURATION = 3 * 60; // load 3 minutes of models each load cycle
const int LOAD_MODEL_BUFFER_TIME = 15;         // we need to load the models ahead of time (in seconds), it (unfortunately) also means that when we get a reload models request (caused by clock searching eg) during these 15 seconds, we will actually have to wait out this period before the new model will take effect

class Mark5B_Control : public Thread, public State_Machine, public SU_Controller {
  friend class SU_Connection;
  friend class SU_Command_Channel;
  friend class Transparent_channel; 

 public:
  Mark5B_Control(Thread_ID my_id, Thread_ID parent, Thread_Master *owner, int su, bool evlbi, SU_Command_Channel *comm1, Transparent_channel* comm2);
  virtual ~Mark5B_Control();

  static int construct_bit_stream_mask(Corr_Job_Descr* cjd, Logical_Array* array, Logical_Station* station, Scan* scan, Ext_Configure_Scan* conf_scan_msg = NULL);

 protected:
  Mark5B_Control();

  enum States { send_init_messages = 0, connect, initialize, idle, wait_for_cjd, reinitialize, wait_for_initialize, get_polynomials, wait_for_configure, wait_for_data, wait_for_ROT, wait_for_job_starting, configure_next_scan, running, aborting};
  States _current_state;
  std::string state_to_string(States s);
  void set_current_state(States new_state);

  enum DOM_State_Bit { HEADER_ERR = 1, DCM0 = 2, DCM1 = 4, SDRAM_CLK_STOPPED = 8, SDRAM_INIT_DONE = 256, SDRAM_FILL = 512+1024, SDRAM_BUFFER_EMPTY = 2048, SCAN_CONFIGURED = 0x40000000, SPOOLING = 0x80000000 };

  Mark5B_Connection _mark5b_connection;
  Msg_Handler_ID _controller_id;

#ifndef MK5B_TEST
  Object_Store _store;
  Polynomials_DB _poly_lib;
  HW_SW_Setup_DB _hw_sw_setup_lib;
#endif

  std::string get_mark5_ip();
  std::string _mark5_ip;
  int _su_id;

  bool _eVLBI;
  Disk_Unit _disk_unit; // to send commands to remote station only

  bool do_abort();
  bool exit_process();

  Signal_Flag _signal_flag;
  int _key_stop_su_control; //for trapping ^C

  Mark5B_Status _du_status;
  bool handle_enquire(SU_Control_Enquiry_Aux* aux_table);

#ifndef MK5B_TEST
  SU_Command_Channel* _su_cmd; //link to su_cmd_channel object
  Transparent_channel* _transp_channel;
#endif

  void reset_timer();
  bool check_timeout(float period);
  bool check_timeout(); // == check_timeout(_timeout_period);
  void set_timeout_period(float period); // set timeout period and resets timer
  NiceTimer _timer;
  int _timer_key;
  float _timeout_period;

  int _job_id;
  Corr_Job_Descr* _cjd;
  T_Job_Descriptor* _tcjd;

  Long_Time_Value _time_stamp; // mark all message with the time set on construction of this object
  Long_Time_Value _local_time; // updated by the function set_local_time
  void set_local_time();

  bool send_initialize_msg(int bocf_rate);
  bool send_lookup_scan(const std::string& scan_name);
  bool configure_scan();
  bool send_configure_scan_msg(); // precondition: valid _header_contents
  bool send_model_data_msg();
  bool send_pcal_request_msg();
  bool send_ready_msg(); // informs processor control that this mk5 is ready
  void send_positioned_msg();
  bool configure_e_station();
  void send_resync_request(ieee_dbl before); // sends a message to processeing job, requesting a correlator resync before "before"
  void finish_subjob(); // sends a message to the medium handlers that we are finished with this subjob and close connections

  ieee_dbl rot2tot_offset(); // domino needs to buffer some data, this means that depending on the data rate, the rot should be certain time behind the tot, return value in 32e6/s

  Scan_Iterator _scan_iterator;

  bool delete_msg_on_cmd_queue(Forward_List_Iter<Incoming_Message_Info*>* iter);

  Polys* get_polys();
  Polys* _polys;
  ieee_dbl _polynomial_begin_ROT;
  ieee_dbl _poly_offset; // the offset that should be applied to the delay poly when send to to the SU, used for on the fly clock searching, in seconds, the offset with which the first coefficient of the delay polynomials shall be altered


  void do_exit_to_idle(unsigned int error_code, unsigned int action_code = 0, unsigned int ref_id = 0);
  void transmit_error(unsigned int error_code, unsigned int action_code = 0, unsigned int ref_id = 0);
  SU_Control_Error _error_msg;

  inline void set_msg_time_stamps(CCC_Msg& msg);

  bool send_alarm_request(unsigned int parameter, ieee_dbl obey_rot);
  bool check_alarm();

  Scan_Bounds* _scan_bounds;
  Header_Contents* _header_contents; // before configuring a scan, its bytes start has to be found, so before entering running and after each scan configuration, we send a request for the next scan, the result is stored here

  Logical_Array* get_logical_array(ieee_dbl start, ieee_dbl end);

  unsigned int _output_mapping[NUM_CHANNELS];
  bool init_output_mapping(Corr_Job_Descr* cjd);

 private:
  // SU_Controller virtual function
  virtual unsigned int get_suid_code() const ;
  virtual unsigned int get_job_id() const ;
  virtual Msg_Handler_ID& get_SUC_ID() ;
  virtual bool initiated() const ;
  virtual int transmit_tape_status(unsigned int status = 0, unsigned int error_code = 0) ;
  virtual int transmit_disk_status(unsigned int status = 0, unsigned int error_code = 0) ;
  virtual bool is_mounted() const { return _du_status.get_mounted_vsn() != ""; }
  virtual const char* get_barcode(int bank) const { _du_status.get_mounted_vsn(bank).c_str(); }
  virtual void get_head_positions(int_32 head_pos[MAX_HEADS_PER_DPU]) const { } // does not apply to mark5B
  virtual int_32 get_DPU_id() const { return 0; } // does not apply to mark5B

  virtual bool get_allocation() const ; // returns true iff the dpu is has a medium mounted

  virtual bool queue_response(Incoming_Message_Info* info); // should not be called for mark5b

  // state machine virtual functions
  virtual int do_state();
  virtual int do_transition() { return 0; }
  virtual int find_state() { return 0; }
  virtual int find_transition() { return 0; }
  virtual int pre_state();
  virtual int pre_transition() { return 0; }

  // thread virtual functions
  virtual int run();
  virtual int shutdown();
  virtual int install();

  static const int GOOD_DATA_TIMEOUT = 60; // wait maximum of 60s for good (valid headers) data
  static const int START_TIME = 120; // give the system 120s to set up a rot clock, after being configured

};

// inlines

void Mark5B_Control::set_msg_time_stamps(CCC_Msg& msg) {
  set_local_time();
  msg.set_local_time(_local_time);  
  msg.set_time_stamp(_time_stamp); 
}
  
#endif
