#include <su_control/mark5b_connection.h>

#include <iostream>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 

using namespace std;

Mark5B_Connection::Mark5B_Connection() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONNECTION,"Mark5B_Connection()" << endl);
  _read_pointer = _end_of_read_buffer = &_read_buffer[0];
  _connected = false;
  _client_socket = -1;
}

Mark5B_Connection::~Mark5B_Connection() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONNECTION,"~Mark5B_Connection()" << endl);
  if (_client_socket >= 0) {
    close(_client_socket);
  }
  // FIX delete content of _received_messages
  NDBG(LOGSTREAM,DEBUG|CONNECTION,"~Mark5B_Connection() done" << endl);  
}

bool Mark5B_Connection::connect_socket(const string& host) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONNECTION,"connect_socket(const string& host)" << endl);
  // try once/s
  static time_t last_time = 0;
  time_t now = time(NULL);
  if (now - last_time <= 0) {
    return false;
  }

  last_time = now;
  if (_client_socket < 0) {
    _client_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (_client_socket < 0) {
      NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR opening socket" << endl);
      return false;
    }
  }
  
  struct hostent* hostent = gethostbyname(host.c_str());
  if (hostent == NULL) {
    NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR, no such host" << endl);
    return false;
  }

  struct sockaddr_in serv_addr;
  memset((char *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  memcpy((char*)&serv_addr.sin_addr.s_addr, (char*)hostent->h_addr, hostent->h_length);

  struct servent* servent = getservbyname("m5drive", "tcp");
  if (servent == NULL) {
    NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR, cannot find port number" << endl);
  }
  serv_addr.sin_port = servent->s_port;
  if (connect(_client_socket,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) {
    NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR connecting" << endl);
    perror("Socket connect");
    return false;
  }
  _connected = true;
  return true;
}

bool Mark5B_Connection::connected() { 
  return _client_socket >= 0 && _connected; 
}

int Mark5B_Connection::retreive_messages(timeval& max_wait, int su_id) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONNECTION,"retreive_messages(timeval& max_wait, int su_id)" << endl);
  if (_client_socket == -1 || !_connected) {
    NDBG(LOGSTREAM,WARNING|CONNECTION,"Socket not (open or connected)" << endl);
    return -2;
  }
  fd_set active_sockets; // set of sockets waiting to be read, as reported by select
  FD_ZERO(&active_sockets);
  FD_SET(_client_socket, &active_sockets);
  //NDBG(LOGSTREAM,CONNECTION|DEBUG,"Table size: " << getdtablesize() << endl);
  select(getdtablesize(), &active_sockets, NULL, NULL, &max_wait);
  
  int old_nr_messages = _received_messages.size();
  while (_client_socket >= 0 && FD_ISSET(_client_socket, &active_sockets)) {
    //NDBG(LOGSTREAM,CONNECTION|DEBUG,"Got something on socket " << _client_socket << endl);
    struct header header; // first get the header of the message then decide which message to receive
    if (receive_msg(&_client_socket,&header)) {
      ((struct Binary_Message*)_message_buffer)->bin_head = header;

      // if exactly enough bytes for the header are ready on the socket, a next read call would block, so do another select (this assumes message contain more bytes than a header)
      fd_set more_bytes;
      FD_ZERO(&more_bytes);
      FD_SET(_client_socket,&more_bytes);
      max_wait.tv_sec = 0;
      max_wait.tv_usec = 0;
      if (select(getdtablesize(), &more_bytes, NULL, NULL, &max_wait) < 1) {
	NDBG(LOGSTREAM,WARNING|CONNECTION,"Expected more bytes than just a header" << endl);
	return -1;
      }
      NDBG(LOGSTREAM,NOISE|CONNECTION,"Received header msg code " << hex << header.msg_code << dec << endl);
      switch (header.msg_code) {
      case (MOUNT_CHANGE) :
	receive_msg_content(&_client_socket, (struct Mount_Change*)_message_buffer, su_id);
	break;
      case (M5_STATUS) :
	receive_msg_content(&_client_socket, (struct M5_Status*)_message_buffer, su_id);
	break;
      case (SCAN_BOUNDS) :
	receive_msg_content(&_client_socket, (struct Scan_Bounds*)_message_buffer, su_id);
	break;
      case (HEADER_CONTENTS) :
	receive_msg_content(&_client_socket, (struct Header_Contents*)_message_buffer, su_id);
	break;
      case (PCAL_DATA) :
        receive_msg_content(&_client_socket, (struct Pcal_Data*)_message_buffer, su_id);
	break;
      case (SCOUNT_DATA) :
	receive_msg_content(&_client_socket, (struct Scount_Data*)_message_buffer, su_id);
	break;
      default:
	NDBG(LOGSTREAM,WARNING|CONNECTION,"Unknown message received " << header.msg_code << ", clearing socket buffer" << endl);
	// clear the socket, to get in a clean state
	int in;
	do {
	  in = 0;
	  FD_ZERO(&active_sockets);
	  FD_SET(_client_socket, &active_sockets);
	  max_wait.tv_sec = 0;
	  max_wait.tv_usec = 0;
	  select(getdtablesize(), &active_sockets, NULL, NULL, &max_wait);
	  if (_client_socket >= 0 && FD_ISSET(_client_socket, &active_sockets)) {
	    in = read_chunk(&_client_socket,MAX_MESSAGE_SIZE);
	    NDBG(LOGSTREAM,DEBUG|CONNECTION,"Throwing away " << in << " bytes" << endl);
	  }
	  _read_pointer = _end_of_read_buffer = &_read_buffer[0]; // get rid of the bytes 
	} while (in == MAX_MESSAGE_SIZE && _client_socket >= 0);

      } // end switch

    } // if header in
    _read_pointer = &_read_buffer[0];
    _end_of_read_buffer = &_read_buffer[0];

    if (_client_socket >= 0) {
      FD_ZERO(&active_sockets);
      FD_SET(_client_socket, &active_sockets);
      max_wait.tv_sec = 0;
      max_wait.tv_usec = 0;
      select(getdtablesize(), &active_sockets, NULL, NULL, &max_wait);
    }
  } // while socket has bytes waiting
  return _received_messages.size() - old_nr_messages;
}

// private/protected

Mark5B_Connection::Message_Storage::iterator Mark5B_Connection::find_message(int msg_code) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONNECTION,"Mark5B_Connection::find_message(int msg_code)" << endl);
  Message_Storage::iterator msg = _received_messages.begin(); 
  while (msg != _received_messages.end() && ((header*)*msg)->msg_code != msg_code) {
    NDBG(LOGSTREAM,NOISE|CONNECTION,"Find message: passing msg code " << ((header*)*msg)->msg_code << endl);
    msg++;
  }
  return msg;
}

// read messages to buffer

int Mark5B_Connection::read_chunk(int* client_socket, int number_of_bytes) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONNECTION,"read_chunk(int* client_socket, int number_of_bytes)" << endl);
  int bytes_left = _end_of_read_buffer - _read_pointer;
  if (bytes_left < 0) {
    NDBG(LOGSTREAM,CRITICAL|CONNECTION,-bytes_left << " uninitialized bytes out of buffer used" << endl);
  }
  if (number_of_bytes - bytes_left > 0) {
    //NDBG(LOGSTREAM,DEBUG|CONNECTION,"Trying to read " << number_of_bytes - bytes_left << " bytes from socket " << *client_socket << endl);
    int in = read(*client_socket, _end_of_read_buffer, number_of_bytes - bytes_left);
    NDBG(LOGSTREAM,NOISE|CONNECTION,"Read returned " << in << " out of " << number_of_bytes - bytes_left << ", bytes_left=" << bytes_left << endl);
  
    if (in < 0) {
      NDBG(LOGSTREAM,CRITICAL|CONNECTION,"ERROR reading from socket" << endl);
      perror("Socket read");
      return -1;
    }
    if (in == 0) {
      close(*client_socket);
      *client_socket = -1;
      _connected = false;
      NDBG(LOGSTREAM,LOG|CONNECTION,"Closed socket connection" << endl);
      return 0;
    }
    _end_of_read_buffer += in;
  }
  return _end_of_read_buffer - _read_pointer;
}

void Mark5B_Connection::read_value(char* buffer, int number_of_bytes) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONNECTION,"read_value(char* buffer, int number_of_bytes)" << endl);
  int count;
#ifdef FIX_ENDIAN
  for (count = number_of_bytes-1; count >= 0; count--) {
#else
  for (count = 0; count < number_of_bytes; count++) {
#endif
    //NDBG(LOGSTREAM,DEBUG|CONNECTION,"Read from buffer pos: " << _read_pointer - &_read_buffer[0] << " max: " << MAX_MESSAGE_SIZE << endl);
    buffer[count] = *_read_pointer;
    _read_pointer++;
  }
}

void Mark5B_Connection::write_value(const char* buffer, int number_of_bytes) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONNECTION,"write_value(const char* buffer, int number_of_bytes)" << endl);
  int count;
#ifdef FIX_ENDIAN
  for (count = number_of_bytes-1; count >= 0; count--) {
#else
  for (count = 0; count < number_of_bytes; count++) {
#endif
    *_write_pointer = buffer[count];
    _write_pointer++;
  }
}
