#include <su_control/mark5b_control.h>
#include <su_control/message_defs.h>

#include <hw_sw_setup_db/su_params.h>
#include <hw_sw_setup_db/evlbi_params.h>

#include <jccs_util/comp_mediumid.h>

#include <fundamental/ccc_constants.h>
#include <map>

using namespace std;

const int MICRO_WAIT = 100000;
const int DEFAULT_BOCF_RATE = 32;

Mark5B_Control::Mark5B_Control(Thread_ID my_id, Thread_ID parent, Thread_Master *owner, int su, bool evlbi, SU_Command_Channel *comm1, Transparent_channel* comm2 ) :  
  Thread(my_id, parent, owner), 
#ifndef MK5B_TEST
#ifdef USE_MYSQL
  _store(Database_Access::mySQL),
#endif
#ifdef USE_MSQL1
  _store(Database_Access::mSQL1),
#endif
  _poly_lib(_store),
  _hw_sw_setup_lib(_store),
#endif
  _timer(&_signal_flag) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"Mark5B_Control(Thread_ID my_id, Thread_ID parent, Thread_Master *owner, int su, SU_Command_Channel *comm1 )" << endl);

#ifndef MK5B_TEST
  _poly_lib.initialise();
  _hw_sw_setup_lib.initialise();  
  _su_cmd = comm1; // takes care of communication with rest of JCCS (not mk5b)
  _transp_channel = comm2;
#endif

  _su_id = su;
  _mark5_ip = get_mark5_ip();
  _key_stop_su_control = _signal_flag.register_for_signal(SIGINT);
  _timer_key = _timer.register_for_timer();

  _tcjd = NULL;
  _scan_bounds = NULL;
  _header_contents = NULL;

  _job_id = 0;

  _eVLBI = evlbi;

  set_current_state(send_init_messages);
}

Mark5B_Control::~Mark5B_Control() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"~Mark5B_Control()" << endl);
  _signal_flag.deregister_from_signal(SIGINT,_key_stop_su_control);
  delete _tcjd;
  delete _scan_bounds;
  delete _header_contents;
  set_stat(Thread::terminating);
  _su_cmd->quit();
  _transp_channel->quit();
  NDBG(LOGSTREAM,DEBUG|CONTROL,"~Mark5B_Control() done" << endl);
}

unsigned int Mark5B_Control::get_suid_code() const {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"get_suid_code()" << endl);
  return _su_id;
}

unsigned int Mark5B_Control::get_job_id() const {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"get_job_id()" << endl);
  return _job_id;
}

Msg_Handler_ID& Mark5B_Control::get_SUC_ID() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"get_SUC_ID()" << endl);
  return _controller_id;
}

bool Mark5B_Control::initiated() const {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"initiated()" << endl);
  return true;
}

int Mark5B_Control::transmit_tape_status(unsigned int status, unsigned int error_code) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"transmit_tape_status(unsigned int status, unsigned int error_code)" << endl);
  NDBG(LOGSTREAM,WARNING|CONTROL,"Legacy function transmit_tape_status called" << endl);
  return 0;
}

bool Mark5B_Control::queue_response(Incoming_Message_Info* info) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"queue_response(Incoming_Message_Info* info)" << endl);
  NDBG(LOGSTREAM,WARNING|CONTROL,"Legacy function queue_response called" << endl);
}

bool Mark5B_Control::get_allocation() const {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"get_allocation()" << endl);
  // define the (constant) mapping
  typedef pair<States, bool> Map_Element;
  typedef map<Map_Element::first_type, Map_Element::second_type> Mapping;
  const Map_Element map_elements[] = { make_pair(send_init_messages, false),
                                       make_pair(connect,false), 
				       make_pair(initialize,false),
				       make_pair(idle,false),
				       make_pair(wait_for_cjd, true),
				       make_pair(reinitialize, true),
				       make_pair(wait_for_initialize, true),
				       make_pair(get_polynomials, true),
				       make_pair(wait_for_configure, true),
				       make_pair(wait_for_data, true),
				       make_pair(wait_for_ROT, true),
				       make_pair(wait_for_job_starting, true),
                                       make_pair(configure_next_scan, true),
				       make_pair(running, true),
				       make_pair(aborting, false) };

  const Mapping mapping(map_elements, map_elements + sizeof(map_elements)/sizeof(map_elements[0]));

  Mapping::const_iterator res = mapping.find(_current_state);
  if (res != mapping.end()) {
    return res->second;
  }
  else {
    NDBG(LOGSTREAM, WARNING|CONTROL, "get_allocation state is not defined " << _current_state << endl);
    return false;
  }
}

int Mark5B_Control::pre_state() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"pre_state()" << endl);

  // check for interrupt
  if ((_signal_flag.signal_occurred(SIGINT, _key_stop_su_control))) {
    NDBG(LOGSTREAM,INFO|CONTROL,"Received ^C. Exiting Process" << endl);
    exit_process();
    return 0;
  }

  // check for connection problems
  if (_current_state > connect && !_mark5b_connection.connected() && _current_state != aborting) { // lost connection
    NDBG(LOGSTREAM,WARNING|CONTROL,"Lost socket connection to mark5, trying to reconnect\n");
    if (!_mark5b_connection.connect_socket(_mark5_ip)) {
      NDBG(LOGSTREAM,WARNING|CONTROL,"Connection repair failed, returning to connect state");
      set_current_state(connect);
    }
  }
  
  // update disk unit
  if (_eVLBI) {
    Init_Proc::get_du_parameters(_store,_disk_unit,_eVLBI,_su_id);
    if (_disk_unit.du_status.get_VSN(0) != _du_status.get_mounted_vsn() && _current_state == idle) {
      _du_status.set_mounted_vsn(_disk_unit.du_status.get_VSN(0),0);
      transmit_disk_status();
    }
  }

  // get messages from socket
  timeval wait = {0,MICRO_WAIT};
  _mark5b_connection.retreive_messages(wait, _su_id);
  
  // check for mount_change
  Mount_Change* mount_msg;
  while (mount_msg = _mark5b_connection.get_message<Mount_Change>()) {
    NDBG(LOGSTREAM,CONTROL|LOG,"Received mount change, bank A: \'" << string(mount_msg->vsn[0],8) << "\' bank B \'" << string(mount_msg->vsn[1],8) << "\'" << endl);
    if (_current_state == initialize) { // first mount msg received after default initialize
      set_current_state(idle);
    }
    else if (_current_state == reinitialize) { // first mount msg received after initialize with the actual BOCF
      set_current_state(get_polynomials);
    }
    if (!_eVLBI) {
      _du_status.process_mounted_msg(*mount_msg);
      transmit_disk_status();
    }
    delete mount_msg;
  }

  // check for m5_status
  M5_Status* m5_status;
  while (m5_status = _mark5b_connection.get_message<M5_Status>()) {
    NDBG(LOGSTREAM,CONTROL|LOG,"Received status message, status word: " << hex << m5_status->status << dec << endl);
    _du_status.process_m5_status(*m5_status);
    delete m5_status;

    if (_current_state == wait_for_initialize) { // if reply to configure is in goto get_polys
      // there is no way of knowing whether this message is caused by the configure message we send in reinitialize or any previous configuration command, but it is all we have
      set_timeout_period(0); // try to get polys immediately
      set_current_state(get_polynomials);
    }

    if (_current_state == wait_for_data) {
      if (_du_status.get_status_word() & SPOOLING) { // the first good data has arrived
	// msg proc job we are ready for the ROT clock
	send_positioned_msg();
	set_current_state(wait_for_ROT);
      }
      else if (check_timeout(GOOD_DATA_TIMEOUT)) {
	NDBG(LOGSTREAM, CONTROL|CRITICAL, "No valid headers received after " << GOOD_DATA_TIMEOUT << "s, aborting" << endl);
	do_abort();
      }
    }

    // if domino sets its header error bit and we are allocated to a job, domino has been aborted, so we should abort our process to
    if ( (_du_status.get_status_word() & HEADER_ERR) && get_allocation() ) {
      NDBG(LOGSTREAM, CONTROL|WARNING, "Header error detected, aborting" << endl);
      do_abort();
    }

  }
  
  // check for scan bound message
  Scan_Bounds* scan_bounds;
  while (scan_bounds = _mark5b_connection.get_message<Scan_Bounds>()) {
    if (_scan_bounds) {
      NDBG(LOGSTREAM,WARNING|CONTROL,"Double Scan Bounds message received" << endl);
      delete scan_bounds;
    }
    else {
      NDBG(LOGSTREAM,DEBUG|CONTROL,"Scan Bounds received" << endl);
      // bank selected, find the header now
      // first delete old response

      delete _header_contents;
      _header_contents = NULL;

      _scan_bounds = scan_bounds;
      Ext_Find_Next_Header find_header(get_job_id());
      find_header.byte_num = (long long)_scan_iterator.get_current_scan()->get_start_footage(); // no scan_bounds->start_byte since lookup scan is only used as a bank select
      _mark5b_connection.send_message(find_header);
    }
  }
  
  // check for Header Contents
  Header_Contents* header;
  while (header = _mark5b_connection.get_message<Header_Contents>()) {
    if (_header_contents) {
      NDBG(LOGSTREAM,WARNING|CONTROL,"Header_Contents message received, while Header_Contents are already know" << endl);
      delete header;
    }
    else {
      _header_contents = header;

      //print header found
      NDBG(LOGSTREAM,CONTROL|DEBUG,"Header found:" << endl << "byte_num " << header->byte_num << endl << "df_header " << hex << endl);
      for (int dim0 = 0; dim0 < sizeof(header->df_header)/sizeof(header->df_header[0]); dim0++) {
	NDBG(LOGSTREAM,CONTROL|DEBUG,header->df_header[dim0] << " ");
      }
      NDBG(LOGSTREAM,CONTROL|DEBUG, endl << dec << "su_id " << (int)header->su_id << endl << endl);


      // we have to configure the scan with a whole second header to start (else mk5b gives an header error), so keep looking if this one is not a whole second
      Ext_Find_Next_Header find_header(get_job_id());
      find_header.byte_num = _header_contents->byte_num;
      if (find_header.byte_num < 0) {
        NDBG(LOGSTREAM, WARNING|CONTROL, "Unable to find header, aborting" << endl);
        do_abort();
      }
      else {
        int frame_nr = _header_contents->df_header[1] & 0x7fff; // frame# within second
        if (frame_nr != 0) {
          // from the data rate we can calculate where we would expect the header to be

          //get speedup and nr tracks in scan
          int data_rate; //10e6 b/s
          Logical_Array* array = get_logical_array(_cjd->get_sub_job_starting_ROT().get_sysclks(), _cjd->get_sub_job_ending_ROT().get_sysclks());
          Record_Format* record_format = _scan_iterator.get_current_scan()->get_record_format();
          if (array && record_format) {
            double muxmode = record_format->get_fan_out_factor();
            int num_bits = record_format->get_number_of_bits_sampling();
            double speedup = array->get_playback_speedup_rate();
            int nr_tracks_in_scan = array->get_nr_of_channels() * (int)muxmode * num_bits;
            data_rate = nr_tracks_in_scan * (int)round(_scan_iterator.get_current_scan()->get_data_rate());          }
          else {
            NDBG(LOGSTREAM, WARNING|CONTROL, "Failed to find logical array entry or record format, guessing data rate" << endl);
            // ok, we can't get the data rate, but we still can look for the header, just assume the lowest possible data rate, which means small steps to find the header
            data_rate = 32;
          }
          int frames_per_second = data_rate / 8 * 100; // convert to byte/s (/8 * 1000000), 1 frame is 10000 bytes

          NDBG(LOGSTREAM,DEBUG|CONTROL,"Data rate " << data_rate << " frame nr " << frame_nr << " frames/s " << frames_per_second << endl);
          
          find_header.byte_num += (frames_per_second - frame_nr) * 10016; // frame has 10000 data bytes + 16 bytes header

          delete _header_contents;
          _header_contents = NULL;
          _mark5b_connection.send_message(find_header);
        }
        else {
          // we found a whole second header, configure
          if (!configure_scan()) { // send configure scan msg to mk5, start lookup of next scan
            NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to configure scan, aborting" << endl);
            do_abort();
          }
          else if (_current_state != running) { // if this is the first scan we configure for this subjob
            send_positioned_msg();
            set_current_state(wait_for_ROT);
          }
        }
      } // else (some header found)
    } // not double header contents
  } // while header content in queue

  // check for new poly ids (result of clock searching)
  while(in_cmd_queue(CCC_Action_Request::Poly_Pid)) {
    NDBG(LOGSTREAM,DEBUG|CONTROL,"New poly pid received" << endl);
    Command_Message* cmd = (Command_Message*)cmd_queue.remove(cmd_iter)->get_msg();
    Poly_Pid_Table* poly_pid_table = ((Poly_Pid_Table*)(((General_CCC_Table&)(cmd->get_table(SU_Table_Types::Poly_Pid))).get_table_data()));
    Logical_Station* current_station = _scan_iterator.get_containing_station();
    if (current_station) {
      NDBG(LOGSTREAM,LOG|CONTROL,"Replacing polys for " << current_station->get_station_ID() << endl);
      current_station->set_persistent_polys_id(poly_pid_table->get_poly_pid());
      get_polys();
      
      send_model_data_msg();
    }
    else {
      NDBG(LOGSTREAM,CRITICAL|CONTROL,"Failed to get station to set new polynomials, aborting");
      do_abort();
    }
    delete cmd;
  }
	
  while (this->in_cmd_queue(CCC_Action_Request::Poly_Offset)) {
    Command_Message* msg = (Command_Message*)cmd_queue.remove(cmd_iter)->get_msg();
    _poly_offset = ((Poly_Offset_Table*)(((General_CCC_Table&)(msg->get_table(SU_Table_Types::Poly_Offset))).get_table_data()))->get_offset();
    NDBG(LOGSTREAM, WARNING|CONTROL, "!!!Warning!!! poly offset set to " << _poly_offset << endl);
    send_model_data_msg();
    delete msg;
  }


  if (in_cmd_queue(CCC_Action_Request::Abort)) {
    NDBG(LOGSTREAM,LOG|CONTROL,"Abort request received in pre_state" << endl);
    do_abort();
    delete_msg_on_cmd_queue(cmd_iter);
  }

  return 0;
}

int Mark5B_Control::do_state() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"do_state() " << _current_state << endl);
  switch (_current_state) {
  case (send_init_messages) : {
    send_ready_msg(); // message to PC
    transmit_disk_status(); // let su_monitor know we are here    
    if (_mark5_ip != "") {
      set_current_state(connect);
    }
    else {
      set_current_state(aborting);
    }
    break;
  }
  case (connect) : {
    if (_mark5b_connection.connect_socket(_mark5_ip)) {
      set_timeout_period(0); // send init msg immediate
      set_current_state(initialize);
    }
    break;
  }
  case (initialize) : {
    // domino will reply with a mount change message to indicate succesful receiving init msg, this is handled in pre_state
    if (check_timeout()) {
      send_initialize_msg(DEFAULT_BOCF_RATE);
    }
    break;
  }
  case (idle) : {
    // wait for enquire msg
    if (in_cmd_queue(CCC_Action_Request::Disconn_SU_Control)) {
      delete_msg_on_cmd_queue(cmd_iter);
      exit_process();
    }
    if (in_cmd_queue(CCC_Action_Request::Enquire_SU_Control)) {
      SU_Control_Enquiry_Aux* aux_table = ((SU_Control_Enquiry_Aux *)((General_CCC_Table &)((Command_Message *)cmd_iter->contents()->get_msg())->get_table(SU_Table_Types::CCC_SUC_Enquire)).get_aux_data());
		
      if(!_du_status.enquire_vsn(aux_table->get_bar_code()))  {
	NDBG(LOGSTREAM,WARNING|CONTROL, "Actual Disk_id[0] = " << _du_status.get_mounted_vsn(0) << endl <<  "Actual Disk_id[1] = " << _du_status.get_mounted_vsn(1) << endl << "Requested Disk_id = " << aux_table->get_bar_code() << endl);
	
      }
      else {
	unsigned int op_code = (unsigned int)(aux_table->get_op_code());
	//for now pay attention to tape operation codes only
	op_code = op_code & 0XFF;
		
	switch (op_code) {
	case  Medium_Operation_Medium_Ops::Processing_Job : {
	  if (handle_enquire(aux_table)) {
	    set_current_state(wait_for_cjd);
	  }
	  break;
	}
	case Medium_Operation_Medium_Ops::Stop : {
	  NDBG(LOGSTREAM,WARNING|CONTROL,"Received Medium_Operation_Medium_Ops::Stop enquire command while in idle, sending abort command, but nothing should be playing" << endl);
	  do_abort();
	  break;
	}
	default :
	  handle_enquire(aux_table);
	  transmit_disk_status();
	  break;
	} // switch op_code
      } // else (if !vsn enquired)
      delete_msg_on_cmd_queue(cmd_iter);
    } // if enquire in queue
    break;
  }
  case (wait_for_cjd) : {
    // wait for cjd in msg queue -> init mk5b

    while(this->action_req_in_cmd_queue(SU_Action_Values::CJD)) {
      delete _tcjd; // deletes the CJD inside to
      _tcjd = (T_Job_Descriptor *)cmd_iter->contents()->get_msg();
      _cjd = (Corr_Job_Descr *)_tcjd->get_job_descriptor();
      delete cmd_queue.remove(cmd_iter);
      if (_cjd->get_corr_job_ID() != get_job_id()) {
	do_exit_to_idle(SU_Control_Error_Codes::Job_ID_Exception, SU_Action_Values::CJD, (unsigned int)_tcjd->get_msg_id());
      }
      else {
	// initialize scan iterator with cjd
	if (!_scan_iterator.set_cjd(_cjd,_du_status.get_mounted_vsn())) {
	  do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
	  break;
	}
	else {
	  if (_scan_iterator.get_current_scan()) {
	    set_timeout_period(0); // send init msg immediately
	    set_current_state(reinitialize);
	  }
	  else {
	    NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to set current scan: no scans to correlate, back to idle" << endl);
	    do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
	    break;
	  }
	}

	if (!init_output_mapping(_cjd)) {
	  do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
	}

      } // else (cjd has our job_id)
    } // while cjd in msg queue
    break;
  }
  case (reinitialize) : {
    //send initialize message to mark5b (this time using the correct bocf) until a mount change has been sent as a reply (caugth in prestate), then send a positioned message
    if (check_timeout()) {
      send_initialize_msg(_cjd->get_bocf_rate());
      set_current_state(wait_for_initialize); // set state to wait for reply to this command
    }
    break;
  }
  case (wait_for_initialize) : {
    break; // handled in prestate
  }
  case (get_polynomials) : {
    if (check_timeout()) {
      if (_polys = get_polys()) {
        // for eVLBI we load the model in chunks, _polynomial_begin_ROT will represent the begin of the chunk loaded with send_model_data_msg
        _polynomial_begin_ROT = _cjd->get_sub_job_starting_ROT().get_sysclks();
        // we can introduce an artificial offset to look for fringes on the fly
        _poly_offset = 0;

	if (send_model_data_msg() && send_pcal_request_msg()) {
	  NDBG(LOGSTREAM,LOG|CONTROL,"Model data and phase cal request send to mk5" << endl);
	}
	else {
	  NDBG(LOGSTREAM,LOG|CONTROL,"Failed to send model data and/or phase cal request to mk5" << endl);	  
	}
	if (_eVLBI) {
	  if (!configure_scan()) { // send configure scan msg to mk5
            NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to configure scan, aborting" << endl);
            do_abort();
	  }
	  else {
	    // send commands to dimino
	    if (configure_e_station()) {
	      set_current_state(wait_for_data);
	      reset_timer();
	    }
	    else {
	      NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to configure remote station" << endl);
	      do_abort();
	    }
	  }
	}
	else {
	  send_lookup_scan(_scan_iterator.get_current_scan()->get_scan_ID()); // this will initiate the configure procedure, lookup scan is only used as a bank select
	  set_current_state(wait_for_configure);
	}
      }
      else {
	NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to get polynomials, retrying" << endl);
	set_timeout_period(5);
      }
    }
    break;
  }
  case (wait_for_configure) : 
  case (wait_for_data) :  {
    break; // handled in prestate
  }
  case (wait_for_ROT) : {
    while(in_cmd_queue(CCC_Action_Request::SU_Control_Set_ROT)) {
      // construct medium ready msg
      Short_Time_Value time(_cjd->get_sub_job_starting_ROT().get_sysclks());
      Su_Control_Medium_Ready msg(_su_id, time, time, _output_mapping);
      msg.set_sender_id(get_SUC_ID());
      msg.set_ref_id(0); //unsolicited?
      set_msg_time_stamps(msg);
#ifndef MK5B_TEST
      _su_cmd->transmit_medium_handlers(&msg);
#endif

      set_current_state(wait_for_job_starting);
      delete_msg_on_cmd_queue(cmd_iter);

    }

    break;
  }
  case (wait_for_job_starting) : {
    //check for arrival of job_starting cmd
    if(in_cmd_queue(CCC_Action_Request::SU_Control_Job_Strt)) {
      //get data handler service id 
      int serv_type = (int)((Job_Starting_Aux *)((General_CCC_Table &)((Command_Message *)cmd_iter->contents()->get_msg())->get_table(SU_Table_Types::CCC_SUC_Job_Starting)).get_aux_data())->get_data_handler_service_id();
#ifndef MK5B_TEST
      if(!(_su_cmd->open_DATA_HANDLER_Connection(serv_type))) {
	NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to open connection to data handler" << endl);
	do_exit_to_idle(SU_Control_Error_Codes::Data_Handler_Conn_Exception);
      }
      else {
	set_current_state(running);
      }
#endif
      delete_msg_on_cmd_queue(cmd_iter);

      // the first alarm request will need to be send after the ROT clock is started, do it here
      if (!_eVLBI) {
	NDBG(LOGSTREAM,DEBUG|CONTROL,"Load scan alarm sent with ROT " << Short_Time_Value(_scan_iterator.get_end_scan()->get_stop_TOT()).get_rot_string() << endl);
	// in case of eVLBI, gaps have been removed in the configure scan message
	send_alarm_request(ROT_Clock_Alarm_Codes::Load_Scan + _su_id, _scan_iterator.get_end_scan()->get_stop_TOT());
      }
      else {
        cout << "Poly begin rot " << Short_Time_Value(_polynomial_begin_ROT).get_rot_string() << endl;
        cout << "Offsetting " << Short_Time_Value(SYSTEM_RATE * (EVLBI_LOAD_MODEL_DURATION - LOAD_MODEL_BUFFER_TIME)).get_rot_string() << endl;
        send_alarm_request(ROT_Clock_Alarm_Codes::Load_Polys + _su_id, _polynomial_begin_ROT + SYSTEM_RATE * (EVLBI_LOAD_MODEL_DURATION - LOAD_MODEL_BUFFER_TIME));
      }
    }
    break;
  }
  case (configure_next_scan) : {
    if ((_du_status.get_status_word() & SCAN_CONFIGURED) == 0) { // wait for abort to complete
      // lookup next scan
      Scan* scan = _scan_iterator.get_current_scan();
      if (scan) {
        if (_eVLBI) {
          NDBG(LOGSTREAM, WARNING|CONTROL, "Received a scan load request in eVLBI mode: logic error");
        }
        else {
          send_lookup_scan(scan->get_scan_ID());
        }
        if (send_model_data_msg() && send_pcal_request_msg()) {
          NDBG(LOGSTREAM,LOG|CONTROL,"Model data and phase cal request send to mk5" << endl);
        }
        else {
          NDBG(LOGSTREAM,LOG|CONTROL,"Failed to send model data and/or phase cal request to mk5" << endl);
        }
        //send_resync_request(scan->get_start_TOT()); not required, it seems
        set_current_state(running);
      }
      else {
        NDBG(LOGSTREAM,WARNING|CONTROL,"No next scan in configure next scan" << endl);
        do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
      }
    }
    else {
      NDBG(LOGSTREAM, DEBUG|CONTROL, "Waiting for domino to confirm abort before configuring next scan" << endl);
    }
    break;
  }
  case (running) : {
    // check for alarm to load next scan, load polys for that scan 
    if (!check_alarm()) {
      do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
    }

    Pcal_Data* pcal_data;
    while (pcal_data = _mark5b_connection.get_message<Pcal_Data>()) {
      NDBG(LOGSTREAM,DEBUG|CONTROL,"Discarding phase cal data message" << endl);
    }
    break;
  }
    
  default:
    NDBG(LOGSTREAM,CRITICAL|CONTROL,"Unrecognized state in do_state " << _current_state << endl);
    break;
  }
  while (_mark5b_connection.get_message<Mount_Change>() || 
	 _mark5b_connection.get_message<M5_Status>() || 
	 _mark5b_connection.get_message<Scan_Bounds>() || 
	 _mark5b_connection.get_message<Header_Contents>() || 
	 _mark5b_connection.get_message<Pcal_Data>() || 
	 _mark5b_connection.get_message<Scount_Data>()) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Unprocessed message from mark5" << endl);
  }
  return 0;
}


int Mark5B_Control::run() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"run()" << endl);
  return cycle();
}


int Mark5B_Control::shutdown() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"shutdown()" << endl);
  return 0;
}


int Mark5B_Control::install() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"install()" << endl);
  return 0;
}



int main(int argc, char** argv) {
  if (argc != 3) {
    cout << "wrong number of parameters, expected 2, got " << (argc - 1) << endl;
    return 2;
  }
  //assume PROCESSOR_CONTROL starts up some object SU_CONTROL with the su_id parameter
   
   cout << "su_control: argc = " << argc << endl;
   cout << "su_control: argv1 = " << argv[1] << endl;
   cout << "su_control: argv2 = " << argv[2] << endl;
   int su = atoi(argv[1]);
   bool evlbi = atoi(argv[2]);
   
   Thread_Master master;
   dbg_process_id = "SU_CONTROL";

#ifndef MK5B_TEST
   // the master takes ownership
   SU_Command_Channel* comm1 = new SU_Command_Channel(master.new_thread_id(),0, &master, su);
   Transparent_channel* comm2 = new Transparent_channel(master.new_thread_id(),0, &master, comm1, su);
   Mark5B_Control* control = new Mark5B_Control(master.new_thread_id(), 0, &master, su, evlbi, comm1, comm2);

   //initalise all channels, etc
   if(!(comm1->initialise(control)) || !(comm2->initialise(control, NULL /* used to be comm3*/))) {
     return 2;
   }
#else
   Mark5B_Control* control = new Mark5B_Control(master.new_thread_id(), 0, &master, su, evlbi, NULL, NULL);
#endif

   master.add_thread(control);
#ifndef MK5B_TEST
   master.add_thread(comm1);
   master.add_thread(comm2);
#endif

   cout << "running" << endl;
   while(! master.run()) {
   }
   return 0;
}

