#include <su_control/mark5b_control.h>
#include <su_control/message_defs.h>

#include <hw_sw_setup_db/su_params.h>
#include <hw_sw_setup_db/evlbi_params.h>

#include <jccs_util/comp_mediumid.h>

#include <errors/texts.h>
#include <errors/errors.h>

#include <math.h>
#include <bitset>

using namespace std;

const int INITIALIZE_TIMEOUT = 5;

string Mark5B_Control::get_mark5_ip() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"get_mark5_ip()" << endl);
#ifndef MK5B_TEST
  SU_Params_Query query = *((SU_Params_Query *)_store.get_query_object(SU_PARAMS));
  query.set_su_id(_su_id);
  query.set_su_id_cond(Table_Column::eq);
  query.set_assoc_du_ip_cond(Table_Column::any);
  Iter_List<Query_Object *>* query_result = _store.get_multi_object(&query);
      
  if(! query_result) {
    if (persistent_error == 0)  {
      NDBG(LOGSTREAM,CRITICAL|CONTROL,"No matches to SU query" << endl);
    }
    else {
      NDBG(LOGSTREAM,CRITICAL|CONTROL,"Query failed" << endl);
    }
  }
  else {
    Forward_List_Iter<Query_Object *> result_iter(query_result);
    SU_Params_Query *result = (SU_Params_Query *)result_iter.first(); 
    return result->get_assoc_du_ip();
  } 
  return "";
#else
  return "10.88.0.58";
#endif
}

string Mark5B_Control::state_to_string(States s) {
  switch (s) {
  case (send_init_messages) : {
    return "send init messages";
  }
  case (connect) : {
    return "connect";
  }
  case (initialize) : {
    return "initialize";
  }
  case (idle) : {
    return "idle";
  }
  case (wait_for_cjd) : {
    return "wait_for_cjd";
  }
  case (reinitialize) : {
    return "reinitialize";
  }
  case (wait_for_initialize) : {
    return "wait_for_initialize";
  }
  case (get_polynomials) : {
    return "get_polynomials";
  }
  case (wait_for_data) : {
    return "wait_for_data";
  }
  case (wait_for_configure) : {
    return "wait_for_configure";
  }
  case (wait_for_ROT) : {
    return "wait_for_ROT";
  }
  case (wait_for_job_starting) : {
    return "wait_for_job_starting";
  }
  case (configure_next_scan) : {
    return "configure_next_scan";
  }
  case (running) : {
    return "running";
  }
  case (aborting) : {
    return "aborting";
  }
  default :
    return "'no string defined for state'";
  }
}

void Mark5B_Control::set_current_state(States new_state) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"set_current_state(States new_state)" << endl);
  if (_current_state != new_state) {
    NDBG(LOGSTREAM,LOG|CONTROL,"State transition " << state_to_string(_current_state) << "  -> " << state_to_string(new_state) << endl);
    _current_state = new_state;
    transmit_disk_status();
  }
}

bool Mark5B_Control::do_abort() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"do_abort()" << endl);
  Aborted_Action_Status msg(_su_id, 
			    0, // status
			    _job_id);
  set_msg_time_stamps(msg);
  msg.set_sender_id(get_SUC_ID());

#ifndef MK5B_TEST
  _su_cmd->transmit_medium_handlers(&msg);
  _su_cmd->close_medium_handler_conns();
  _su_cmd->close_data_handler_conn();
#endif
  if (_mark5b_connection.connected()) {
    set_current_state(initialize);
  }
  else {
    set_current_state(connect);
  }
  if (_eVLBI) {
    _disk_unit.in2net("disconnect");
    _disk_unit.in2net("off");
  }
  return _mark5b_connection.send_message(Ext_Abort_Scan(get_job_id()));
}

bool Mark5B_Control::exit_process() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"exit_process()" << endl);
  do_abort();
  transmit_disk_status();
#ifndef MK5B_TEST
  _su_cmd->quit();
#endif
  set_stat(Thread::terminating);
  get_controller()->shutdown();
}

void Mark5B_Control::reset_timer() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"reset_timer()" << endl);
  _timer.reset(_timer_key);
  if (!_timer.running()) {
    _timer.start_timer();
  }
}

bool Mark5B_Control::check_timeout(float period) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"check_timeout(float period)" << endl);
  if (!_timer.running()) {
    _timer.start_timer();
  }
  return _timer.elapsed_time(_timer_key) >= period;
}

bool Mark5B_Control::check_timeout() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"check_timeout()" << endl);
  return check_timeout(_timeout_period);
}

void Mark5B_Control::set_timeout_period(float period) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"set_timeout_period(float period)" << endl);
  _timeout_period = period;
  reset_timer();
}

void Mark5B_Control::set_local_time() {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"set_local_time()" << endl);
  static struct timeval tod;
  static struct timezone tzp;
  ::gettimeofday(&tod, &tzp);
 
   tm *temp = ::gmtime(&tod.tv_sec);
   ieee_dbl sysclks = SYSTEM_RATE * ((temp->tm_yday*24 + temp->tm_hour)* 60 + temp->tm_min)*60 + temp->tm_sec;

   _local_time.set_sysclks(sysclks);
   // y2k windowing hack... valid until 2089...
   _local_time.set_base_date( (temp->tm_year<90) ? (2000 + temp->tm_year) : (1900 + temp->tm_year) );
}

bool Mark5B_Control::handle_enquire(SU_Control_Enquiry_Aux* aux_table) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"handle_enquire(SU_Control_Enquiry_Aux* aux_table)" << endl);
  Connection *target = cmd_iter->contents()->get_conn();
  if (!target) {
    return false;
  }

  // parameters to send in reply
  unsigned int status = 0;
  if (_du_status.get_mounted_vsn() != "") {
    status = SU_Control_SU_Status_Codes::Medium_Mounted;
  }
  if (aux_table->get_allocate_reqd() && status) {
    _job_id = aux_table->get_job_id();
    status |=  SU_Control_SU_Status_Codes::SU_Allocated;
  }
  char medium_id[MAX_NR_BANKS][MEDIUM_ID_LEN];
  for (int bank = 0; bank < MAX_NR_BANKS; bank++) {
    strncpy(medium_id[bank],_du_status.get_mounted_vsn(bank).c_str(),MEDIUM_ID_LEN);
  }
  SU_Control_Status msg(_su_id, medium_id, DPUType, get_DPU_id(), status, get_job_id());
  set_msg_time_stamps(msg);
  msg.set_ref_id(((CCC_Msg *)cmd_iter->contents()->get_msg())->get_msg_id());
  msg.transmit(*target);
  return status & SU_Control_SU_Status_Codes::SU_Allocated;
}

bool Mark5B_Control::send_initialize_msg(int bocf_rate) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_initialize_msg(int bocf_rate)" << endl);
  Ext_Initialize_Mk5 init_msg(get_job_id());
  init_msg.cf_length = (int)SYSTEM_RATE / bocf_rate;
  if (_cjd && _scan_iterator.get_current_scan()) {
    // copied the setting of bocf_strech from init_proc_p.C
    if (_cjd->get_feature_bits(5).empty() ) {
      init_msg.bocf_stretch = _scan_iterator.get_current_scan()->get_record_format()->get_record_oversampling_factor();
    }
    else {
      init_msg.bocf_stretch = 1;
    }
    init_msg.bocf_stretch = (U8)(init_msg.bocf_stretch * SYSTEM_RATE / _scan_iterator.get_current_scan()->get_record_format()->get_record_bit_rate()); // copied this from init_proc configuration
  }
  else {
    init_msg.bocf_stretch = 1;
  }

  init_msg.su_id = _su_id;
  init_msg.send_mounts = 1; // send all mount changes
  init_msg.realtime = (_eVLBI ? 1 : 0); 
  init_msg.packet_size = _disk_unit.du_status.get_mtu_size();
  if (!_mark5b_connection.send_message(init_msg)) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to send initialize message" << endl);
    return false;
  }
  set_timeout_period(INITIALIZE_TIMEOUT);

  Ext_Send_M5_Status status_request(get_job_id());
  status_request.interval = 1; // request a status every second
  return _mark5b_connection.send_message(status_request);
}

void Mark5B_Control::send_positioned_msg() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_positioned_msg()" << endl);
  Short_Time_Value time(_cjd->get_sub_job_starting_ROT().get_sysclks());
  Short_Time_Value eVLBI_clock_delay(rot2tot_offset());
  Su_Control_Medium_Positioned msg(_su_id, eVLBI_clock_delay, time);
  NDBG(LOGSTREAM,DEBUG|CONTROL,"sending positioned msg: " << endl << msg << endl);
  msg.set_sender_id(get_SUC_ID());
  msg.set_ref_id(_tcjd->get_msg_id());
  set_msg_time_stamps(msg);
#ifndef MK5B_TEST
  _su_cmd->transmit_medium_handlers(&msg);
#endif
}

void Mark5B_Control::send_resync_request(ieee_dbl before) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_resync_request()" << endl);
  Command_Message resync_cmd(CCC_Action_Request::Resync);
  General_CCC_Table* gen_table = new General_CCC_Table(SU_Table_Types::Resync_Request, new Resync_Request_Table(before), NULL);
  resync_cmd.add_table(gen_table);
#ifndef MK5B_TEST
  _su_cmd->transmit_medium_handlers(&resync_cmd);
#endif
 
}

ieee_dbl Mark5B_Control::rot2tot_offset() {
  const Scan* scan = _scan_iterator.get_current_scan();
  if (!scan) {
    NDBG(LOGSTREAM,CONTROL|CRITICAL, "Cannot find scan to compute rot2tot offset, aborting" << endl);
    do_abort();
    return 0;
  }
  Record_Format* record_format = scan->get_record_format();
  if(!record_format) {
    NDBG(LOGSTREAM,CONTROL|CRITICAL, "Scan has no record format, unable to compute rot2tot offset, aborting" << endl);
    do_abort();
    return 0;
  }
  double muxmode = record_format->get_fan_out_factor();
  int num_bits = record_format->get_number_of_bits_sampling();
  //get speedup and nr tracks in scan
  int nr_tracks_in_scan;
  const Logical_Array* array = get_logical_array(_cjd->get_sub_job_starting_ROT().get_sysclks(), _cjd->get_sub_job_ending_ROT().get_sysclks());
  if (array) {
    nr_tracks_in_scan = array->get_nr_of_channels() * (int)muxmode * num_bits;
  }
  else {
    NDBG(LOGSTREAM, CONTROL|CRITICAL,"Cannot find a logical array to compute desired rot2tot offset: aborting");
    do_abort();
    return 0;
  }

  int data_rate = (nr_tracks_in_scan * (int)round(scan->get_data_rate())) / 8; // we want data rate in MB/s io Mb/s

  // domino needs to fill its buffer to play back, there are 3 levels of buffers: the buffer of the io card, the streamstor FIFO and a software buffer, we need to delay the realtime clock from the wall clock to be able to fill these buffers
  double delay_seconds = 
    DOM_BUFF_MB/data_rate + // we need to fill the io buff fully
    MBB_CHUNK_FACTOR/1e6 + // and 2 software chunks are used as a work buffer between the input from net to the output to the fifo
    (4*MBB_SIZE/128e6 - 2*MBB_CHUNK_FACTOR/1e6)/2 + // we want to be in the middle of the remaining buffer at maximum data rate (128MBps = 1Gbps)
    FIFO_LAT; // domino assumes a latency between the fifo and the io card, we want to compensate for that (assumed) latency

  NDBG(LOGSTREAM, DEBUG|CONTROL, "data rate " << data_rate << " so offset(s) " << delay_seconds << endl);
  return SYSTEM_RATE * delay_seconds;

}

bool Mark5B_Control::delete_msg_on_cmd_queue(Forward_List_Iter<Incoming_Message_Info*>* iter) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"delete_msg_on_cmd_queue(Forward_List_Iter<Incoming_Message_Info*>* iter)" << endl);
  Incoming_Message_Info* info = cmd_queue.remove(iter);
  delete info->get_msg();
  delete info;
}

Polys* Mark5B_Control::get_polys(void) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"get_polys(void)" << endl);
#ifndef MK5B_TEST
  Polys *polys;
  unsigned int poly_pid = (unsigned int)(_scan_iterator.get_containing_station()->get_persistent_polys_id());
  if((polys = (Polys *)_store.get_object(POLYNOMIALS, poly_pid))) {
    return polys;
  }
  else {
    NDBG(LOGSTREAM,WARNING|CONTROL,"WARNING: did not get polys" << endl);
    return 0;
  }
#else
  return NULL;
#endif
}

bool Mark5B_Control::send_alarm_request(unsigned int parameter, ieee_dbl obey_rot) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_alarm_request(unsigned int parameter, ieee_dbl obey_rot), rot: " << Short_Time_Value(obey_rot).get_rot_string() << " job id " << get_job_id() << " parameter " << hex << parameter << dec << endl);
  if (_current_state <= wait_for_ROT) {
    // FIX oops;
  }
  //create a default Msg_Handler_ID for the target ie ROT_Clock
  Msg_Handler_ID target_handle;
  Set_Alarm_Request acm(get_SUC_ID(), target_handle, get_job_id(), parameter);
  acm.set_obey_ROT(obey_rot);
  acm.set_flags(SU_Action_Request_Flags::Obey_ROT_Set);
  set_msg_time_stamps(acm);
			
#ifndef MK5B_TEST
  Connection *target = _su_cmd->get_connection(ROT_CLOCK);
  if(target) {
    acm.transmit(*target);
    return true;
  }
  else {
    do_exit_to_idle(SU_Control_Error_Codes::ROT_Clock_Conn_Exception);
    return false;
  }
#else
  return true;
#endif
}

bool Mark5B_Control::configure_scan() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"configure_scan()" << endl);
  bool succes = true;
  if (_eVLBI || (_scan_bounds && _header_contents)) {
    if (send_configure_scan_msg()) {
      NDBG(LOGSTREAM,LOG|CONTROL,"Scan configuration message send for scan " << _scan_iterator.get_current_scan()->get_scan_ID() << endl);
    }
    else {
      NDBG(LOGSTREAM,CRITICAL|CONTROL,"Failed to send scan configuration messages for scan " << _scan_iterator.get_current_scan()->get_scan_ID() << endl);
      succes = false;
    }
  }
  else {
    NDBG(LOGSTREAM,CRITICAL|CONTROL,"No reply to lookup scan or find next header, unable to configure next scan" << endl);
  }

  // the first alarm request will need to be send after the ROT clock is started, so dont do the first here
  if (!_eVLBI && _current_state == running) {
    NDBG(LOGSTREAM,DEBUG|CONTROL,"Load scan alarm sent with ROT " << Short_Time_Value(_scan_iterator.get_end_scan()->get_stop_TOT()).get_rot_string() << endl);
    // in case of eVLBI, gaps have been removed in the configure scan message
    send_alarm_request(ROT_Clock_Alarm_Codes::Load_Scan + _su_id, _scan_iterator.get_end_scan()->get_stop_TOT());
  }
  return succes;
}

bool Mark5B_Control::check_alarm(void) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"check_alarm(void)" << endl);
  bool succes = true;
  while(this->action_req_in_cmd_queue(SU_Action_Values::Alarm)) {
    Alarm_Clock_Message *acm = (Alarm_Clock_Message *) cmd_iter->contents()->get_msg();
    Clock_Message_Entry *entry = acm->first_entry();
    
    do {
      NDBG(LOGSTREAM,CONTROL|DEBUG,"Alarm: " << entry->get_SU_array_code() << " " << hex << entry->get_parameter() << dec << endl);
      if (entry->get_SU_array_code() == get_job_id()){
	if (entry->get_parameter() == ROT_Clock_Alarm_Codes::Load_Scan + _su_id) {
	  _scan_iterator.goto_next_scan();
          if (_scan_iterator.get_current_scan()) {
	    _mark5b_connection.send_message(Ext_Abort_Scan(get_job_id()));
            set_current_state(configure_next_scan);
          }
          else {
            finish_subjob();
          }
	}
	if (entry->get_parameter() == ROT_Clock_Alarm_Codes::Load_Polys + _su_id) {
          _polynomial_begin_ROT += EVLBI_LOAD_MODEL_DURATION * SYSTEM_RATE;
          send_model_data_msg();
          send_alarm_request(ROT_Clock_Alarm_Codes::Load_Polys + _su_id, _polynomial_begin_ROT + SYSTEM_RATE * (EVLBI_LOAD_MODEL_DURATION - LOAD_MODEL_BUFFER_TIME));
        }
	else if (entry->get_parameter() == ROT_Clock_Alarm_Codes::Job_End) {
	  NDBG(LOGSTREAM,LOG|CONTROL,"Job end received: aborting" << endl);
	  do_abort();
	}
      }
    } while (entry = acm->next_entry(entry));
    
    delete_msg_on_cmd_queue(cmd_iter);
  }

  // check for end of finish rot clocks in eVLBI mode (in disk mode we will get a end of scan alarm)
  while(this->action_req_in_cmd_queue(SU_Action_Values::Finish_ROT)) {
    Finish_ROT_Clock_Message* fcm = (Finish_ROT_Clock_Message *) cmd_iter->contents()->get_msg();
    delete cmd_queue.remove(cmd_iter); //msg is not deleted!
    if (_eVLBI) {
      //iterate through its entries
      Clock_Message_Entry *entry = fcm->first_entry();
      int i;
      for (i = 0; i < fcm->how_many() && (entry->get_SU_array_code() != _cjd->get_corr_job_ID()); ++i) {
        entry = fcm->next_entry(entry);
      }
      
      if(i < fcm->how_many()) {
        finish_subjob();
      }
    }
    delete fcm;
  }

  return succes;
}

bool Mark5B_Control::send_lookup_scan(const string& scan_name) { 
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_lookup_scan(const string& scan_name), " << scan_name << endl);
  if (_du_status.get_bank() < 0 || _du_status.get_bank() > 1) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Bank selected for lookup scan is illegal " << _du_status.get_bank() << endl);
    return false;
  }
  Ext_Lookup_Scan lookup(get_job_id());
  strlcpy(lookup.scan_label, scan_name.c_str(), sizeof(lookup.scan_label)/sizeof(lookup.scan_label[0]));
  lookup.bank_select = _du_status.get_bank();
  NDBG(LOGSTREAM,DEBUG|CONTROL,"lookup up scan " << scan_name << " in bank " << (int)lookup.bank_select << endl);
  if (!_mark5b_connection.send_message(lookup)) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to send lookup scan message" << endl);
    return false;
  }
  // clean previous response
  delete _scan_bounds;
  _scan_bounds = NULL;
  delete _header_contents;
  _header_contents = NULL;
  return true;
}

void Mark5B_Control::do_exit_to_idle(unsigned int error_code, unsigned int action_code, unsigned int ref_id) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"do_exit_to_idle(unsigned int error_code, unsigned int action_code, unsigned int ref_id)" << endl);
  transmit_error(error_code, action_code, ref_id);
#ifndef MK5B_TEST
  _su_cmd->close_medium_handler_conns();
  _su_cmd->close_data_handler_conn();
#endif
  set_current_state(idle);
  static Errors errors;
  NDBG(LOGSTREAM,WARNING|CONTROL,"aborting configuration, problem: " << errors.get_error_text(error_code) << " returned to idle state." << endl);

  if (_eVLBI) {
    _disk_unit.in2net("disconnect");
    _disk_unit.in2net("off");
  }

}

void Mark5B_Control::transmit_error(unsigned int error_code, unsigned int action_code, unsigned int ref_id) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"transmit_error(unsigned int error_code, unsigned int action_code, unsigned int ref_id)" << endl);
#ifndef MK5B_TEST
  _error_msg.set_error_code(error_code);
  _error_msg.set_job_id(get_job_id());
  _error_msg.set_action_code(action_code);
  _error_msg.set_su_id(_su_id);
  _error_msg.set_sender_id(get_SUC_ID());
  _error_msg.set_receiver_id(_su_cmd->get_PC_Handle_ID());
  _error_msg.set_ref_id(ref_id);
  set_msg_time_stamps(_error_msg);

  Connection *target = _su_cmd->get_connection(PROCESSOR_CONTROL);
  if(target) _error_msg.transmit(*target);
   
  _su_cmd->transmit_medium_handlers(&_error_msg);
#endif
}   

int Mark5B_Control::transmit_disk_status(unsigned int status, unsigned int error_code) {
  NDBG(LOGSTREAM,METHOD_ENTRY_OFTEN|CONTROL,"transmit_disk_status(unsigned int status, unsigned int error_code)" << endl);
  char station_id[1][TELESCOPE_ID_LEN];
  memset(station_id[0], '\0', TELESCOPE_ID_LEN );
  
  if (_scan_iterator.get_containing_station()) {
    _scan_iterator.get_containing_station()->get_station_ID().copy(station_id[0], TELESCOPE_ID_LEN);
  }
   
  if(error_code == SU_Control_Error_Codes::SU_Connection_Exception) 
    status |= SU_Control_SU_Status_Codes::DPU_Not_Present;
  
  switch (_current_state) {
  case (wait_for_cjd):
    status |= SU_Control_SU_Status_Codes::SU_Allocated;
    break;
  case (reinitialize) :
  case (wait_for_initialize) :
  case (get_polynomials) :
  case (wait_for_configure) :
    status |= SU_Control_SU_Status_Codes::Configuring_SU;
    status |= SU_Control_SU_Status_Codes::SU_Allocated;
    break;
  case (wait_for_ROT) :
  case (wait_for_job_starting) :
    status |= SU_Control_SU_Status_Codes::Configuring_SU;
    status |= SU_Control_SU_Status_Codes::Heads_Peaked;
    status |= SU_Control_SU_Status_Codes::Medium_Waiting;
    status |= SU_Control_SU_Status_Codes::SU_Allocated;
    break;
  case (running) :
    status |= SU_Control_SU_Status_Codes::Heads_Peaked;
    status |= SU_Control_SU_Status_Codes::Medium_Running;
    status |= SU_Control_SU_Status_Codes::Medium_Processing;
    status |= SU_Control_SU_Status_Codes::SU_Allocated;
    break;
  default:
    break;
  }

  char medium_id[MAX_NR_BANKS][MEDIUM_ID_LEN];
  for (int i = 0; i < MAX_NR_BANKS; i++) {
    strncpy(medium_id[i], _du_status.get_mounted_vsn(i).c_str(), MEDIUM_ID_LEN);
  }

  if (is_mounted()) {
    status |= SU_Control_SU_Status_Codes::Medium_Mounted;
  }
   
  // fake heads
  int_32 head_positions[4];
  for(int i = 0; i < 4; ++i) head_positions[i] = 1;

  SU_Control_SU_Status msg(_su_id,
			   medium_id,
			   DPUType,
			   get_DPU_id(),
			   status | SU_Control_SU_Status_Codes::SU_Allocated ? get_job_id() : 0,
			   status,
			   error_code,
			   1, // stations
			   station_id,
			   1, // num head stacks
			   head_positions,
			   true, // use mark5
			   _eVLBI); // eVLBI

  set_msg_time_stamps(msg);
  msg.set_action_code(SU_Action_Values::Unsolicited);
  msg.set_sender_id(get_SUC_ID());
#ifndef MK5B_TEST
  msg.set_receiver_id(_su_cmd->get_SUM_Handle_ID());
  Connection* target = _su_cmd->get_connection(SU_MONITOR);
  if(target) {
    //NDBG(LOGSTREAM,CONTROL|DEBUG,"Sending disk status with medium " << medium_id << " inactive " << &medium_id[strlen(medium_id)+1] << endl);
    msg.transmit(*target);
    return status;
  }
  return -1;
#else
  return status;
#endif
}
 
Logical_Array* Mark5B_Control::get_logical_array(ieee_dbl start, ieee_dbl end) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"get_logical_array(ieee_dbl start, ieee_dbl end)" << endl);
  Action_Table* table = _cjd->get_action_table();
  Logical_Array* array = NULL;
  if(table->reset_la_iterator()) {
    ieee_dbl best_overlap = 0;
    do {
      Begin_End_Entry* entry = table->get_current_la();
      ieee_dbl overlap = min(entry->get_end_ROT(), end) - max(entry->get_begin_ROT(), start);
      if (overlap > best_overlap) {
	array = _cjd->get_la(entry->get_la_id());
        best_overlap = overlap;
      }
    } while(table->get_next_la());
  }
  else {
    NDBG(LOGSTREAM,WARNING|CONTROL,"action table reset failed" << endl);
  }
  return array;
}


bool Mark5B_Control::send_configure_scan_msg() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_configure_scan_msg()" << endl);
  Ext_Configure_Scan conf_scan(get_job_id());
  Record_Format* record_format = _scan_iterator.get_current_scan()->get_record_format();
  if (!record_format) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Cannot get record format for current scan to configure scan" << endl);
    return false;
  }

  if (!_eVLBI) {
    if (_header_contents->byte_num < 0) {
      NDBG(LOGSTREAM,WARNING|CONTROL,"Could not find a valid header" << endl);
      return false;
    }
    // calculate start/end ROT and byte from scan bounds and header contents
    int julian_year = (int)_cjd->get_obs_TOT_start().get_base_date() + 4799;
    int julian_begin_of_year = julian_year * 365 + julian_year / 4 - julian_year / 100 + julian_year / 400 - 31739; // julian day number for the beginning of the observation year
    int julian_day_number = 100 * (_header_contents->df_header[2] >> 28 & 0xf) 
      + 10 * (_header_contents->df_header[2] >> 24 & 0xf) 
      +      (_header_contents->df_header[2] >> 20 & 0xf); // actually % 1000
    int day_of_year = (julian_day_number - julian_begin_of_year % 1000 + 1000) % 1000;
    int seconds_in_day = 10000 * (_header_contents->df_header[2] >> 16 & 0xf)
      + 1000 * (_header_contents->df_header[2] >> 12 & 0xf)
      +  100 * (_header_contents->df_header[2] >>  8 & 0xf)
      +   10 * (_header_contents->df_header[2] >>  4 & 0xf)
      +        (_header_contents->df_header[2]       & 0xf);

    conf_scan.byte_num = _header_contents->byte_num;
    conf_scan.start_rot = (day_of_year*24*60*60 + seconds_in_day + 1) * SYSTEM_RATE; // FIX +1: see calc_dom_times for "correct" solution
  }
  else {
    time_t now = time(0);
    tm* gmt = gmtime(&now);
    conf_scan.start_rot = SYSTEM_RATE * (gmt->tm_yday*24*60*60 + gmt->tm_hour*60*60 + gmt->tm_min*60 + gmt->tm_sec + START_TIME);
  }

  Logical_Array* array = get_logical_array(_cjd->get_sub_job_starting_ROT().get_sysclks(), _cjd->get_sub_job_ending_ROT().get_sysclks());
  if (!array) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"Cannot get logical array for sub job" << endl);
    return false;
  }
  
  // fill the crossbar map with some values recognizable for debug
  for (int i = 0; i < sizeof(conf_scan.crossbar_map)/sizeof(conf_scan.crossbar_map[0]); i++) {
    conf_scan.crossbar_map[i] = 0xff;
  }

  int bit_stream_mask = construct_bit_stream_mask(_cjd, array, _scan_iterator.get_containing_station(), _scan_iterator.get_current_scan(), &conf_scan);

  if (_eVLBI) {
    // just let it run to the end of the subjob
    conf_scan.end_rot = _cjd->get_sub_job_ending_ROT().get_sysclks();
  }
  else {
    conf_scan.end_rot = _scan_iterator.get_end_scan()->get_stop_TOT();
  }
  
  conf_scan.ss_invalid_code = 0x11223344; // standard values (according to mail from Roger Cappallo)
  conf_scan.dim_invalid_code = 0x11223344;

  conf_scan.bits_sample = record_format->get_number_of_bits_sampling();
  conf_scan.sample_rate = (U8)(record_format->get_record_bit_rate()/1e6);
  
  conf_scan.output_rate = (U8) (conf_scan.sample_rate * array->get_playback_speedup_rate());
  conf_scan.num_bitstreams = array->get_nr_of_channels() * conf_scan.bits_sample;
  strlcpy(conf_scan.scan_label, "test", sizeof(conf_scan.scan_label)/sizeof(conf_scan.scan_label[0])); // only for fork mode
  return _mark5b_connection.send_message(conf_scan);
}

bool Mark5B_Control::send_model_data_msg() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_model_data_msg()" << endl);
  if (!_scan_iterator.get_current_scan()) {
    NDBG(LOGSTREAM, WARNING|CONTROL, "No scan to send models for" << endl);
    return false;
  }
  int nr_polys_sent = 0;
  Persistent_Assoc_List*  assoc_list = _polys->get_polynomials();
  for (int i = 0; i < assoc_list->how_many(); ++i) {
    Persistent_Association* object = assoc_list->get_assoc(i);
    Poly_set* poly_set = (Poly_set*)object->get_target();
    ieee_dbl begin_ROT, end_ROT;
    if (_eVLBI) {
      begin_ROT = _polynomial_begin_ROT;
      end_ROT = begin_ROT + SYSTEM_RATE * EVLBI_LOAD_MODEL_DURATION;
    }
    else {
      begin_ROT = _scan_iterator.get_current_scan()->get_start_TOT();
      end_ROT = _scan_iterator.get_end_scan()->get_stop_TOT() ;
    }

    if (!(poly_set->get_begin_rot() > end_ROT || poly_set->get_end_rot() < begin_ROT)) {
      // send this set of polys as model data msg to mark5B
      Ext_Model_Data model_data(get_job_id());
      const int channel_to_copy = 0; // mark5B uses only 1 channel for delay

      memcpy(model_data.delay,&poly_set->get_delay_polynomials()[6 * channel_to_copy],sizeof(model_data.delay));
      model_data.delay[0] -= _poly_offset;
      memcpy(model_data.phase,poly_set->get_phase_polynomials(),sizeof(model_data.phase));

      model_data.rot_interval[0] = poly_set->get_begin_rot();
      model_data.rot_interval[1] = poly_set->get_end_rot();
      
      // get channels from logical station and fill sideband using the channel block
      ieee_dbl block_rot = (poly_set->get_begin_rot() + poly_set->get_end_rot()) / 2; // take the average to search for the logical channels block
      Logical_Station* station = _scan_iterator.get_containing_station();

      int block_index = 0; 
      Logical_Channels_Block* block = station->get_log_chan_block_containing_ROT(block_rot);
      if (!block) {
	NDBG(LOGSTREAM,WARNING|CONTROL,"No logical channel block found for ROT " << Short_Time_Value(block_rot).get_rot_string() << ", required to send model data" << endl);
	return false;
      }

      model_data.nchan = block->get_nr_channels();
      model_data.sidebands = 0;
      for (int channel = 0; channel < block->get_nr_channels(); channel++) {
	if (_cjd->get_channel(block->get_logical_channel(channel).get_channel_reference())->get_sideband() == lower) {
	  model_data.sidebands |= 1<<channel;
	}
      }
	 
      Record_Format* record_format = _scan_iterator.get_current_scan()->get_record_format();
      if (!record_format) {
	NDBG(LOGSTREAM,WARNING|CONTROL,"Cannot get record format for current scan to configure scan" << endl);
	return false;
      }
      model_data.oversampling = record_format->get_record_oversampling_factor();
      model_data.k_div = (U8)round(log(SYSTEM_RATE / record_format->get_record_bit_rate())/log(2));
      
      if (!_mark5b_connection.send_message(model_data)) {
	NDBG(LOGSTREAM,WARNING|CONTROL,"Failed to send model data to mk5" << endl);
      }
      else {
	nr_polys_sent++;
      }
    } // if poly set ROTs in scan range

  } // for all poly sets
  if (nr_polys_sent == 0) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"No polys sent" << endl);
    return false;
  }
  return true;
}

bool Mark5B_Control::send_pcal_request_msg() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_pcal_request_msg()" << endl);
  Ext_Pcal_Request msg(get_job_id());

  msg.ap = 0;
  char* env_value = getenv("PCAL_ACC_PERIOD");
  if (env_value) {
    msg.ap = atoi(env_value);
  }
  if (msg.ap == 0) {
    msg.ap = PCAL_ACC_PERIOD;
  }

  Phase_Cal_Tone_Table* phase_cal_tones = _scan_iterator.get_containing_station()->get_phase_cal_tones();

  ieee_dbl block_rot = (_scan_iterator.get_current_scan()->get_start_TOT() + _scan_iterator.get_current_scan()->get_stop_TOT()) /2 ;
  Logical_Channels_Block* block = _scan_iterator.get_containing_station()->get_log_chan_block_containing_ROT(block_rot);
  if (!block) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"No logical channel block found for ROT " << Short_Time_Value(block_rot).get_rot_string() << ", required to send pcal_request" << endl);
    return false;
  }
  msg.ntones = block->get_nr_channels();
  for (int channel = 0; channel < msg.ntones; channel++) {
    Phase_Cal_Entry* phase_cal_entry = phase_cal_tones->get_entry(block->get_logical_channel(channel).get_channel_reference().c_str());
    msg.tones[channel] = phase_cal_entry->get_frequency(); // FIX it is not this simple? see send_pcal_request in suman code/init_proc
  }

  return _mark5b_connection.send_message(msg);
}

bool Mark5B_Control::send_ready_msg() {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"send_ready_msg()" << endl);
  SU_Control_Ready msg (_su_id, _du_status.get_mounted_vsn().c_str(), (int)is_mounted());
  set_msg_time_stamps(msg);
  msg.set_sender_id(get_SUC_ID());
#ifndef MK5B_TEST
  msg.set_receiver_id(_su_cmd->get_PC_Handle_ID());
  
  Connection *target = _su_cmd->get_connection(PROCESSOR_CONTROL);
  if(target) {
    msg.transmit(*target);
    if(transmit_disk_status() == -1) {
      exit_process();
      return false;
    }
  }
  else {
    exit_process();
    return false;
  }
#endif
  return true;
}

bool Mark5B_Control::init_output_mapping(Corr_Job_Descr* cjd) {
  NDBG(LOGSTREAM,METHOD_ENTRY|CONTROL,"init_output_mapping(Corr_Job_Descr* cjd)" << endl);
  int channels = -1;
  for (int block = 0; block < _scan_iterator.get_containing_station()->get_nr_of_log_chan_blocks(); block++) {
    Logical_Channels_Block* log_chan_block = _scan_iterator.get_containing_station()->get_log_chan_block(block);
    if (log_chan_block->get_begin_ROT() <= _cjd->get_sub_job_starting_ROT() && _cjd->get_sub_job_starting_ROT() <= log_chan_block->get_end_ROT()) {
      channels = log_chan_block->get_nr_channels();
    }
  }
  if (channels < 0) {
    NDBG(LOGSTREAM,WARNING|CONTROL,"No logical channel block found" << endl);
    return false;
  }
  for (int channel = 0; channel < NUM_CHANNELS; channel++) {
    if (channel < channels) { 
      _output_mapping[channel] = channel; // only true if muxmode <= 3, since mark5b does not use tracks anymore, i assume muxmode == 3 (this is all according to Init_Proc)
    }
    else {
      NDBG(LOGSTREAM,DEBUG|CONTROL,"Setting _output_mapping[" <<  channel << "] to " << (unsigned int)0xffffffff << endl);
      _output_mapping[channel] = 0xffffffff;
    }
  }
  return true;
}


bool Mark5B_Control::configure_e_station() {
  // ensure remote mark5 is not playing
  _disk_unit.in2net("disconnect");
  _disk_unit.in2net("off");

  Record_Format* record_format = _scan_iterator.get_current_scan()->get_record_format();
  if(!record_format) {
    NDBG(LOGSTREAM, WARNING|CONTROL,"Scan has no record format" << endl);
    do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
    return false;
  }
  double muxmode = record_format->get_fan_out_factor();
  int num_bits = record_format->get_number_of_bits_sampling();
  //get speedup and nr tracks in scan
  double speedup;
  int nr_tracks_in_scan;
  Logical_Array* array = get_logical_array(_cjd->get_sub_job_starting_ROT().get_sysclks(), _cjd->get_sub_job_ending_ROT().get_sysclks());
  if (array) {
    speedup = array->get_playback_speedup_rate();
    nr_tracks_in_scan = array->get_nr_of_channels() * (int)muxmode * num_bits;
  }
  else {
    NDBG(LOGSTREAM, WARNING|CONTROL, "Failed to find logical array entry for subjob, unable to configure scan" << endl);
    do_exit_to_idle(SU_Control_Error_Codes::Unspecified_Error);
    return false;
  }

  int bit_stream_mask = construct_bit_stream_mask(_cjd, array, _scan_iterator.get_containing_station(), _scan_iterator.get_current_scan());
  int data_rate = nr_tracks_in_scan * (int)round(_scan_iterator.get_current_scan()->get_data_rate());
  int num_bit_streams = bitset<32>(bit_stream_mask).count();
  int clock_rate = 32;
#if 0 // seems like this is something that needs to be set per experiment, not job
  if (!_disk_unit.set_e_station_clock(clock_rate,"int")) return false;
  if (!_disk_unit.set_e_station_dot()) return false;
#endif
  
  int decimation = (clock_rate * num_bit_streams) / (int)(data_rate * speedup);
  if (!_disk_unit.set_e_station_dim_mode("ext", bit_stream_mask, decimation)) return false;

  if (!_disk_unit.set_e_station_protocol(_disk_unit.du_status.get_protocol(), _disk_unit.du_status.get_sockbuf_size(), _disk_unit.du_status.get_workbuf_size())) return false;
  if (!_disk_unit.in2net("connect")) return false;
  return _disk_unit.in2net("on");
}

void Mark5B_Control::finish_subjob() {
  SU_Control_Finished msg(_su_id, CCC_Finish_Codes::Normal_Finish);
  _su_cmd->transmit_medium_handlers(&msg);

  _su_cmd->close_medium_handler_conns();
  _su_cmd->close_data_handler_conn();

  _mark5b_connection.send_message(Ext_Abort_Scan(get_job_id()));
  if (_mark5b_connection.connected()) {
    set_current_state(initialize);
  }
  else {
    set_current_state(connect);
  }
  if (_eVLBI) {
    _disk_unit.in2net("disconnect");
    _disk_unit.in2net("off");
  }
}
